@@ -35,11 +35,7 @@
  *
  * #JsonArray is the representation of the object type inside JSON. It contains
  * #JsonNode<!-- -->s, which may contain fundamental types, arrays or other
- * objects. Each member of an object is accessed using its name. Please note
- * that the member names are normalized internally before being used; every
- * delimiter matching the %G_STR_DELIMITER macro will be transformed into an
- * underscore, so for instance "member-name" and "member_name" are equivalent
- * for a #JsonObject.
+ * objects. Each member of an object is accessed using its name.
  *
  * Since objects can be expensive, they are reference counted. You can control
  * the lifetime of a #JsonObject using json_object_ref() and json_object_unref().
@@ -127,7 +123,9 @@ json_object_unref (JsonObject *object)
     g_atomic_int_compare_and_exchange (&object->ref_count, old_ref, old_ref - 1);
   else
     {
+      g_list_free (object->members_ordered);
       g_hash_table_destroy (object->members);
+      object->members_ordered = NULL;
       object->members = NULL;
 
       g_slice_free (JsonObject, object);
@@ -139,9 +137,11 @@ object_set_member_internal (JsonObject  *object,
                             const gchar *member_name,
                             JsonNode    *node)
 {
-  gchar *name;
+  gchar *name = g_strdup (member_name);
+
+  if (g_hash_table_lookup (object->members, name) == NULL)
+    object->members_ordered = g_list_prepend (object->members_ordered, name);
 
-  name = g_strdelimit (g_strdup (member_name), G_STR_DELIMITERS, '_');
   g_hash_table_replace (object->members, name, node);
 }
 
@@ -164,8 +164,6 @@ json_object_add_member (JsonObject  *object,
                         const gchar *member_name,
                         JsonNode    *node)
 {
-  gchar *name;
-
   g_return_if_fail (object != NULL);
   g_return_if_fail (member_name != NULL);
   g_return_if_fail (node != NULL);
@@ -223,7 +221,7 @@ json_object_set_member (JsonObject  *object,
 void
 json_object_set_int_member (JsonObject  *object,
                             const gchar *member_name,
-                            gint         value)
+                            gint64       value)
 {
   JsonNode *node;
 
@@ -383,7 +381,7 @@ json_object_set_array_member (JsonObject  *object,
  * Convenience function for setting an object @value of
  * @member_name inside @object.
  *
- * The @object will take ownership of the passed #JsonArray
+ * The @object will take ownership of the passed #JsonObject
  *
  * See also: json_object_set_member()
  *
@@ -399,58 +397,11 @@ json_object_set_object_member (JsonObject  *object,
   g_return_if_fail (object != NULL);
   g_return_if_fail (member_name != NULL);
 
-  node = json_node_new (JSON_NODE_ARRAY);
+  node = json_node_new (JSON_NODE_OBJECT);
   json_node_take_object (node, value);
   object_set_member_internal (object, member_name, node);
 }
 
-/* FIXME: yuck */
-#if GLIB_MAJOR_VERSION == 2 && GLIB_MINOR_VERSION < 14
-static void
-get_keys (gpointer key,
-          gpointer value,
-          gpointer user_data)
-{
-  GList **keys = user_data;
-
-  *keys = g_list_prepend (*keys, key);
-}
-
-static void
-get_values (gpointer key,
-            gpointer value,
-            gpointer user_data)
-{
-  GList **values = user_data;
-
-  *values = g_list_prepend (*values, value);
-}
-
-static GList *
-g_hash_table_get_keys (GHashTable *hash_table)
-{
-  GList *retval = NULL;
-
-  g_return_val_if_fail (hash_table != NULL, NULL);
-
-  g_hash_table_foreach (hash_table, get_keys, &retval);
-
-  return retval;
-}
-
-static GList *
-g_hash_table_get_values (GHashTable *hash_table)
-{
-  GList *retval = NULL;
-
-  g_return_val_if_fail (hash_table != NULL, NULL);
-
-  g_hash_table_foreach (hash_table, get_values, &retval);
-
-  return retval;
-}
-#endif /* GLIB_MAJOR_VERSION == 2 && GLIB_MINOR_VERSION < 14 */
-
 /**
  * json_object_get_members:
  * @object: a #JsonObject
@@ -458,17 +409,22 @@ g_hash_table_get_values (GHashTable *hash_table)
  * Retrieves all the names of the members of a #JsonObject. You can
  * obtain the value for each member using json_object_get_member().
  *
- * Return value: a #GList of member names. The content of the list
- *   is owned by the #JsonObject and should never be modified or
- *   freed. When you have finished using the returned list, use
- *   g_list_free() to free the resources it has allocated.
+ * Return value: (element-type utf8) (transfer container): a #GList
+ *   of member names. The content of the list is owned by the #JsonObject
+ *   and should never be modified or freed. When you have finished using
+ *   the returned list, use g_list_free() to free the resources it has
+ *   allocated.
  */
 GList *
 json_object_get_members (JsonObject *object)
 {
+  GList *copy;
+
   g_return_val_if_fail (object != NULL, NULL);
 
-  return g_hash_table_get_keys (object->members);
+  copy = g_list_copy (object->members_ordered);
+
+  return g_list_reverse (copy);
 }
 
 /**
@@ -477,17 +433,23 @@ json_object_get_members (JsonObject *object)
  *
  * Retrieves all the values of the members of a #JsonObject.
  *
- * Return value: a #GList of #JsonNode<!-- -->s. The content of the
- *   list is owned by the #JsonObject and should never be modified
- *   or freed. When you have finished using the returned list, use
- *   g_list_free() to free the resources it has allocated.
+ * Return value: (element-type JsonNode) (transfer container): a #GList of
+ *   #JsonNode<!-- -->s. The content of the list is owned by the #JsonObject
+ *   and should never be modified or freed. When you have finished using the
+ *   returned list, use g_list_free() to free the resources it has allocated.
  */
 GList *
 json_object_get_values (JsonObject *object)
 {
+  GList *values, *l;
+
   g_return_val_if_fail (object != NULL, NULL);
 
-  return g_hash_table_get_values (object->members);
+  values = NULL;
+  for (l = object->members_ordered; l != NULL; l = l->next)
+    values = g_list_prepend (values, g_hash_table_lookup (object->members, l->data));
+
+  return values;
 }
 
 /**
@@ -498,8 +460,8 @@ json_object_get_values (JsonObject *object)
  * Retrieves a copy of the #JsonNode containing the value of @member_name
  * inside a #JsonObject
  *
- * Return value: a copy of the node for the requested object member
- *   or %NULL. Use json_node_free() when done.
+ * Return value: (transfer full) a copy of the node for the requested
+ *   object member or %NULL. Use json_node_free() when done.
  *
  * Since: 0.6
  */
@@ -523,16 +485,7 @@ static inline JsonNode *
 object_get_member_internal (JsonObject  *object,
                             const gchar *member_name)
 {
-  JsonNode *retval;
-  gchar *name;
-
-  name = g_strdelimit (g_strdup (member_name), G_STR_DELIMITERS, '_');
-
-  retval = g_hash_table_lookup (object->members, name);
-
-  g_free (name);
-
-  return retval;
+  return g_hash_table_lookup (object->members, member_name);
 }
 
 /**
@@ -570,7 +523,7 @@ json_object_get_member (JsonObject  *object,
  *
  * Since: 0.8
  */
-gint
+gint64
 json_object_get_int_member (JsonObject  *object,
                             const gchar *member_name)
 {
@@ -715,7 +668,7 @@ json_object_get_string_member (JsonObject  *object,
  *
  * See also: json_object_get_member()
  *
- * Return value: the array inside the object's member
+ * Return value: (transfer none): the array inside the object's member
  *
  * Since: 0.8
  */
@@ -745,7 +698,7 @@ json_object_get_array_member (JsonObject  *object,
  *
  * See also: json_object_get_member()
  *
- * Return value: the object inside the object's member
+ * Return value: (transfer none): the object inside the object's member
  *
  * Since: 0.8
  */
@@ -778,17 +731,10 @@ gboolean
 json_object_has_member (JsonObject *object,
                         const gchar *member_name)
 {
-  gchar *name;
-  gboolean retval;
-
   g_return_val_if_fail (object != NULL, FALSE);
   g_return_val_if_fail (member_name != NULL, FALSE);
 
-  name = g_strdelimit (g_strdup (member_name), G_STR_DELIMITERS, '_');
-  retval = (g_hash_table_lookup (object->members, name) != NULL);
-  g_free (name);
-
-  return retval;
+  return (g_hash_table_lookup (object->members, member_name) != NULL);
 }
 
 /**
@@ -818,36 +764,23 @@ void
 json_object_remove_member (JsonObject  *object,
                            const gchar *member_name)
 {
-  gchar *name;
+  GList *l;
 
   g_return_if_fail (object != NULL);
   g_return_if_fail (member_name != NULL);
 
-  name = g_strdelimit (g_strdup (member_name), G_STR_DELIMITERS, '_');
-  g_hash_table_remove (object->members, name);
-  g_free (name);
-}
+  for (l = object->members_ordered; l != NULL; l = l->next)
+    {
+      const gchar *name = l->data;
 
-typedef struct _ForeachClosure  ForeachClosure;
+      if (g_strcmp0 (name, member_name) == 0)
+        {
+          object->members_ordered = g_list_delete_link (object->members_ordered, l);
+          break;
+        }
+    }
 
-struct _ForeachClosure
-{
-  JsonObject *object;
-
-  JsonObjectForeach func;
-  gpointer data;
-};
-
-static void
-json_object_foreach_internal (gpointer key,
-                              gpointer value,
-                              gpointer data)
-{
-  ForeachClosure *clos = data;
-  const gchar *member_name = key;
-  JsonNode *member_node = value;
-
-  clos->func (clos->object, member_name, member_node, clos->data);
+  g_hash_table_remove (object->members, member_name);
 }
 
 /**
@@ -870,15 +803,18 @@ json_object_foreach_member (JsonObject        *object,
                             JsonObjectForeach  func,
                             gpointer           data)
 {
-  ForeachClosure clos;
+  GList *members, *l;
 
   g_return_if_fail (object != NULL);
   g_return_if_fail (func != NULL);
 
-  clos.object = object;
-  clos.func = func;
-  clos.data = data;
-  g_hash_table_foreach (object->members,
-                        json_object_foreach_internal,
-                        &clos);
+  /* the list is stored in reverse order to have constant time additions */
+  members = g_list_last (object->members_ordered);
+  for (l = members; l != NULL; l = l->prev)
+    {
+      const gchar *member_name = l->data;
+      JsonNode *member_node = g_hash_table_lookup (object->members, member_name);
+
+      func (object, member_name, member_node, data);
+    }
 }
